## segway 1.2.0.dev-r7941 run f79d633a84f111e19f41000c2997d6ab at 2012-04-12 15:50:57.485928
/net/noble/vol2/home/mmh1/arch/Linux/RHEL6/x86_64/bin/gmtkTriangulate -cppCommandOptions "-DCARD_SEG=4 -DCARD_FRAMEINDEX=2000000 -DSEGTRANSITION_WEIGHT_SCALE=1.0 -DCARD_SUBSEG=1" -outputTriangulatedFile traindir.trackspec/triangulation/segway.str.4.1.trifile -strFile traindir.trackspec/segway.str -verbosity 0
/net/noble/vol2/home/mmh1/arch/Linux/RHEL6/x86_64/bin/gmtkEMtrain -base 3 -componentCache F -cppCommandOptions "-DCARD_SEG=4 -DCARD_FRAMEINDEX=2000000 -DSEGTRANSITION_WEIGHT_SCALE=1.0 -DCARD_SUBSEG=1" -deterministicChildrenStore F -dirichletPriors F -fmt1 binary -fmt2 binary -gpr ^0:-1:0 -hashLoadFactor 0.98 -inputMasterFile traindir.trackspec/params/input.master -island T -iswp1 F -iswp2 F -jtFile traindir.trackspec/log/jt_info.txt -lldp 0.001 -lst 100000 -maxEmIters 1 -nf1 2 -nf2 0 -ni1 0 -ni2 2 -objsNotToTrain traindir.trackspec/auxiliary/dont_train.list -obsNAN T -of1 traindir.trackspec/observations/float32.list -of2 traindir.trackspec/observations/int.list -storeAccFile traindir.trackspec/accumulators/acc.0.1.bin -strFile traindir.trackspec/segway.str -triFile traindir.trackspec/triangulation/segway.str.4.1.trifile -trrng 1 -verbosity 0
/net/noble/vol2/home/mmh1/arch/Linux/RHEL6/x86_64/bin/gmtkEMtrain -base 3 -componentCache F -cppCommandOptions "-DCARD_SEG=4 -DCARD_FRAMEINDEX=2000000 -DSEGTRANSITION_WEIGHT_SCALE=1.0 -DCARD_SUBSEG=1" -deterministicChildrenStore F -dirichletPriors T -fmt1 binary -fmt2 binary -hashLoadFactor 0.98 -inputMasterFile traindir.trackspec/params/input.master -island T -iswp1 F -iswp2 F -jtFile traindir.trackspec/log/jt_info.txt -lldp 0.001 -lst 100000 -maxEmIters 1 -nf1 2 -nf2 0 -ni1 0 -ni2 2 -objsNotToTrain traindir.trackspec/auxiliary/dont_train.list -obsNAN T -of1 traindir.trackspec/observations/float32.list -of2 traindir.trackspec/observations/int.list -storeAccFile traindir.trackspec/accumulators/acc.0.0.bin -strFile traindir.trackspec/segway.str -triFile traindir.trackspec/triangulation/segway.str.4.1.trifile -trrng 0 -verbosity 0
/net/noble/vol2/home/mmh1/arch/Linux/RHEL6/x86_64/bin/gmtkEMtrain -base 3 -componentCache F -cppCommandOptions "-DCARD_SEG=4 -DOUTPUT_PARAMS_FILENAME=traindir.trackspec/params/params.0.params.0 -DCARD_FRAMEINDEX=2000000 -DSEGTRANSITION_WEIGHT_SCALE=1.0 -DCARD_SUBSEG=1" -deterministicChildrenStore F -fmt1 binary -fmt2 binary -hashLoadFactor 0.98 -inputMasterFile traindir.trackspec/params/input.master -island T -iswp1 F -iswp2 F -jtFile traindir.trackspec/log/jt_info.txt -llStoreFile traindir.trackspec/likelihood/likelihood.ll -lldp 0.001 -loadAccFile traindir.trackspec/accumulators/acc.0.@D.bin -loadAccRange 0:1 -lst 100000 -maxEmIters 1 -nf1 2 -nf2 0 -ni1 0 -ni2 2 -objsNotToTrain traindir.trackspec/auxiliary/dont_train.list -obsNAN T -of1 traindir.trackspec/observations/float32.list -of2 traindir.trackspec/observations/int.list -outputMasterFile traindir.trackspec/params/output.master -strFile traindir.trackspec/segway.str -triFile traindir.trackspec/triangulation/segway.str.4.1.trifile -trrng nil -verbosity 0
